/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.ws.cxf;

import org.apache.cxf.Bus;
import org.apache.cxf.BusException;
import org.apache.cxf.BusFactory;
import org.apache.cxf.transport.DestinationFactory;
import org.apache.cxf.transport.DestinationFactoryManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Level;
import static org.ow2.jonas.lib.util.Log.setComponentLogLevel;
import org.ow2.jonas.ws.cxf.http.JOnASHTTPTransportFactory;
import org.ow2.util.execution.ExecutionResult;
import org.ow2.util.execution.IExecution;
import org.ow2.util.execution.helper.RunnableHelper;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * Manage all the code required to start a new CXF Bus.
 * @author Guillaume Sauthier
 */
public class BusCreator {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(BusCreator.class);


    /**
     * Initialize Internal CXF Bus.
     * @return a new fully initialized CXF Bus
     */
    public Bus initBus() {

        // Create a new Bus
        // Creation mechanism use the TCCL to access default Bus configuration
        IExecution<Bus> createBusJob = new IExecution<Bus>() {

            public Bus execute() throws Exception {

                // Bus start/stop operations are extra verbose
                Level old = setComponentLogLevel("org.apache.cxf", BasicLevel.LEVEL_ERROR);
                Level old2 = setComponentLogLevel("org.springframework", BasicLevel.LEVEL_ERROR);

                Bus newBus =  BusFactory.newInstance().createBus();

                // Reset log levels
                setComponentLogLevel("org.apache.cxf", old);
                setComponentLogLevel("org.springframework", old2);

                return newBus;
            }

        };

        ExecutionResult<Bus> res = new RunnableHelper<Bus>().execute(Bus.class.getClassLoader(),
                                                                     createBusJob);

        if (res.hasException()) {
            logger.error("Cannot create Bus", res.getException());
        }

        Bus bus = res.getResult();

        try {
            createDestinationFactory(bus);
        } catch (BusException e) {
            logger.error("Cannot create destination factory", e);
        }

        return bus;

    }

    /**
     * Create the EasyBeans destination factory and register the transports.
     * @throws BusException if the factory is not created.
     * @param bus Bus to be configured
     */
    private void createDestinationFactory(final Bus bus) throws BusException {

        JOnASHTTPTransportFactory factory = new JOnASHTTPTransportFactory(bus);

        // transportIds done by Spring ?
        List<String> namespaces = new ArrayList<String>();
        namespaces.add("http://schemas.xmlsoap.org/wsdl/soap/http");
        namespaces.add("http://schemas.xmlsoap.org/soap/http");
        namespaces.add("http://www.w3.org/2003/05/soap/bindings/HTTP/");
        namespaces.add("http://schemas.xmlsoap.org/wsdl/http/");
        namespaces.add("http://cxf.apache.org/transports/http/configuration");
        namespaces.add("http://cxf.apache.org/bindings/xformat");
        factory.setTransportIds(namespaces);

        for (String ns : namespaces) {
            registerTransport(bus, factory, ns);
        }

    }

    /**
     * Utility method to register {@link org.apache.cxf.transport.DestinationFactory} under a namespace.
     * @param bus Bus to be configured
     * @param factory the {@link org.apache.cxf.transport.DestinationFactory} instance
     * @param namespace the key
     * @throws BusException if registration fails.
     */
    private void registerTransport(final Bus bus,
                                   final DestinationFactory factory,
                                   final String namespace) throws BusException {

        DestinationFactoryManager manager = bus.getExtension(DestinationFactoryManager.class);
        manager.registerDestinationFactory(namespace, factory);
    }
}
