/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.ws.cxf;

import org.apache.cxf.Bus;
import org.apache.cxf.BusFactory;
import org.apache.cxf.jaxws.EndpointImpl;
import org.apache.cxf.jaxws.JaxWsServerFactoryBean;
import org.apache.cxf.jaxws.spi.ProviderImpl;
import org.apache.cxf.jaxws.support.JaxWsImplementorInfo;
import org.apache.cxf.jaxws.support.JaxWsServiceFactoryBean;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Level;
import org.ow2.easybeans.server.Embedded;
import org.ow2.jonas.lib.service.AbsServiceImpl;
import static org.ow2.jonas.lib.util.Log.setComponentLogLevel;
import org.ow2.jonas.service.ServiceException;
import org.ow2.jonas.ws.cxf.easybeans.naming.WebServiceRefExtensionListener;
import org.ow2.jonas.ws.cxf.jaxws.CXFWSEndpoint;
import org.ow2.jonas.ws.jaxws.IJAXWSService;
import org.ow2.jonas.ws.jaxws.IWebServiceEndpoint;
import org.ow2.jonas.ws.jaxws.PortMetaData;
import org.ow2.jonas.ws.jaxws.WSException;
import org.ow2.util.ee.builder.webserviceref.WebServiceRefBuilder;
import org.ow2.util.ee.metadata.common.api.struct.IJaxwsWebServiceRef;
import org.ow2.util.ee.metadata.common.api.struct.IJwsWebService;
import org.ow2.util.ee.metadata.war.api.IWarClassMetadata;
import org.ow2.util.event.api.IEventService;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

import javax.naming.Reference;
import javax.naming.NamingException;
import javax.xml.ws.spi.Provider;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * The CXFService class is used to declare CXF as a module in JOnAS.
 *
 * @author Guillaume Sauthier
 */
public class CXFService extends AbsServiceImpl implements IJAXWSService {

    /**
     * This Bus is shared by all clients inside this JVM.
     * This is the default CXF Bus.
     */
    private Bus clientBus;

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(CXFService.class);

    /**
     * The event service used to register listeners.
     */
    private IEventService eventService;

    /**
     * The naming extension listener for EasyBean's java:comp/env building.
     */
    private WebServiceRefExtensionListener listener;

    /**
     * Abstract start-up method to be implemented by sub-classes.
     *
     * @throws org.ow2.jonas.service.ServiceException
     *          service start-up failed
     */
    protected void doStart() throws ServiceException {

        // CXF Bus init is extra verbose, level up the default logging level
        Level old = setComponentLogLevel("org.apache.cxf", BasicLevel.LEVEL_ERROR);
        Level old2 = setComponentLogLevel("org.springframework", BasicLevel.LEVEL_ERROR);

        // Force initialization of the singleton Bus shared by all clients.
        clientBus = BusFactory.getDefaultBus();

        // Reset log levels
        setComponentLogLevel("org.apache.cxf", old);
        setComponentLogLevel("org.springframework", old2);

        // Force usage of the CXF JAX-WS Provider
        System.setProperty(Provider.JAXWSPROVIDER_PROPERTY, ProviderImpl.class.getName());

        logger.info("CXF service started");
    }

    /**
     * Abstract method for service stopping to be implemented by sub-classes.
     *
     * @throws org.ow2.jonas.service.ServiceException
     *          service stopping failed
     */
    protected void doStop() throws ServiceException {

        Level old = setComponentLogLevel("org.apache.cxf", BasicLevel.LEVEL_ERROR);
        Level old2 = setComponentLogLevel("org.springframework", BasicLevel.LEVEL_ERROR);

        // Stop the client bus
        // true means 'wait for complete stop'
        clientBus.shutdown(true);
        clientBus = null;
        BusFactory.setDefaultBus(null);

        setComponentLogLevel("org.apache.cxf", old);
        setComponentLogLevel("org.springframework", old2);

        logger.info("CXF service stopped");
    }

    /**
     * Construct a Reference dedicated to the given @WebServiceRef metadata.
     *
     * @param serviceRef the reference described by the user.
     * @return a JNDI Naming {@link javax.naming.Reference} that can be bound in a JNDI Context.
     */
    public Reference createNamingReference(final IJaxwsWebServiceRef serviceRef) throws NamingException {

        // This could be subclassed for any CXF related specific needs
        WebServiceRefBuilder builder = new WebServiceRefBuilder();

        return builder.build(serviceRef);
    }

    /**
     * Creates a new POJO Web service endpoint from classes metadata.
     *
     * @param metadata the class own metadata (annotation + XML)
     * @param loader the ClassLoader to be used to load classes from the metadata
     * @return A POJO typed IWebServiceEndpoint based on metadata
     */
    public IWebServiceEndpoint createPOJOWebServiceEndpoint(final IWarClassMetadata metadata,
                                                            final ClassLoader loader)
             throws WSException {

        // TODO Try to limit number of buses (ex 1 bus/module instead of 1 bus/endpoint)
        BusCreator creator = new BusCreator();
        Bus bus = creator.initBus();

        String classname = metadata.getJClass().getName().replace('/', '.');

        Class<?> klass = null;
        try {
            klass = loader.loadClass(classname);
        } catch (ClassNotFoundException e) {
            throw new WSException("Impossible to load the WebService class", e);
        }

        IJwsWebService ws = metadata.getJwsWebService();

        String wsdlName = ws.getWsdlLocation();

        // Load the WSDL if possible
        URL url = null;
        if (wsdlName != null) {
            url = loader.getResource(wsdlName);
            if (url == null) {
                try {
                    url = new URL(wsdlName);
                } catch (MalformedURLException ex2) {
                    url = loader.getResource("/" + wsdlName);
                }
            }
        }

        // End of Generic Part -----------

        EndpointImpl ep = createEndpointImpl(bus, klass, url);

        // find the URL pattern to use

        String serviceName = ws.getServiceName();

        String pattern = null;
        if (!"".equals(serviceName)) {
            // default pattern : /{service-name}
            pattern = "/" + serviceName;
        } else {
            // else /{class.name}Service
            pattern = "/" + klass.getSimpleName() + "Service";
        }

        // Creates PortMetaData to hold deployment/runtime information
        // about the web service
        PortMetaData pmd = new PortMetaData();
        pmd.setUrlPattern(pattern);

        logger.debug("PortMetaData [pattern:{0}]", pmd.getUrlPattern());

        // Wraps the Endpoint into our own IWSEndpoint implementation
        CXFWSEndpoint endpoint = new CXFWSEndpoint(ep,
                                                   IWebServiceEndpoint.EndpointType.POJO,
                                                   pmd);

        logger.info("Endpoint {0} inited", klass.getName());


        return endpoint;
    }

    /**
     * Configure CXF internals for the new JAXWS {@link javax.xml.ws.Endpoint}.
     * @param bus the Bus that will be used to support the new endpoint
     * @param klass POJO's class
     * @param wsdlUrl URL where the WSLD can be loaded
     * @return a configured CXF {@link org.apache.cxf.jaxws.EndpointImpl} instance.
     * @throws WSException if it was impossible to create a new instance from the klass parameter
     */
    private EndpointImpl createEndpointImpl(final Bus bus, final Class<?> klass, final URL wsdlUrl)
               throws WSException {

        JaxWsImplementorInfo jaxWsImplementorInfo = new JaxWsImplementorInfo(klass);
        JaxWsServiceFactoryBean jaxWsServiceFactoryBean = new JaxWsServiceFactoryBean(jaxWsImplementorInfo);
        jaxWsServiceFactoryBean.setBus(bus);
        if (wsdlUrl != null) {
            jaxWsServiceFactoryBean.setWsdlURL(wsdlUrl);
        }

        // Create a service with the correct classloader
        ClassLoader old = Thread.currentThread().getContextClassLoader();
        Object implementor = null;
        try {
            Thread.currentThread().setContextClassLoader(klass.getClassLoader());
            // TODO Why do we need to call create here ?
            jaxWsServiceFactoryBean.create();
            implementor = klass.newInstance();
        } catch (IllegalAccessException e) {
            throw new WSException("Cannot access '" + klass.getName() + "'", e);
        } catch (InstantiationException e) {
            throw new WSException("Cannot instantiate '" + klass.getName() + "'", e);
        } finally {
            // reset
            Thread.currentThread().setContextClassLoader(old);
        }

        // Build the endpoint
        JaxWsServerFactoryBean factoryBean = new JaxWsServerFactoryBean(jaxWsServiceFactoryBean);

        return new EndpointImpl(bus, implementor, factoryBean);
    }

    /**
     * Callback invoked when an EventService becomes available.
     * @param eventService the event service
     */
    public void bindEventService(final IEventService eventService) {
        this.eventService = eventService;
        listener = new WebServiceRefExtensionListener();
        eventService.registerListener(listener, Embedded.NAMING_EXTENSION_POINT);
    }

    /**
     * Callback invoked when the used EventService becomes unavalable.
     * @param eventService disposed service
     */
    public void unbindEventService(final IEventService eventService) {
        this.eventService.unregisterListener(listener);
        this.eventService = null;
    }

}
